@@ -4,9 +4,10 @@
 #include <stdio.h>
 #include <ddk/imm.h>
 #include <limits.h>
-#include "cannaapi.h"
+#include "canna.h"
 #include "ut.h"
 #include "apisup.h"
+#include "list.h"
 
 Array Clients;
 Array Context;
@@ -233,12 +234,18 @@ void Req3(CanHeader* ch,int16_t *p1,uint16_t *p2)
     *p2 = Swap2(((Req3_t*)ch)->p2);
 }
 
+//p5free뤳
+uint16_t* Req4r(CanHeader* h,int16_t* p1,uint16_t* p2,uint16_t* p3,uint16_t* p4)
+{
+    Req9(h,p1,(int16_t*)p2,(int16_t*)p3,(int16_t*)p4);
+    return h->Length==sizeof(Req4_t)-sizeof(*h) ? NULL : ((Req4_t*)h)->p5;
+}
+
 //cannawcharʸޥХʸˤ
 //p5free뤳
-char* Req4(Req4_t* q,int16_t* p1,uint16_t* p2,uint16_t* p3,uint16_t* p4)
+char* Req4(CanHeader* h,int16_t* p1,uint16_t* p2,uint16_t* p3,uint16_t* p4)
 {
-    Req9((CanHeader*)q,p1,(int16_t*)p2,(int16_t*)p3,(int16_t*)p4);
-    return ToMb(q->p5);
+    return ToMb(Req4r(h,p1,p2,p3,p4));
 }
 
 void Req5(CanHeader* h,int16_t *p1,uint16_t *p2,int32_t* p3)
@@ -285,15 +292,14 @@ void *Req10(Req10_t* q,int16_t *p1,int16_t *p2,int32_t *p3)
 uint16_t* Req11r(CanHeader* ch,int16_t* p1,int16_t* p2)
 {
     Req3(ch,p1,(uint16_t*)p2);
-    return ((Req11_t*)ch)->p3;
+    return ch->Length==sizeof(*p1)+sizeof(*p2) ? NULL : ((Req11_t*)ch)->p3;
 }
 
 //p3ϣХʸǤȤƣХʸѴ
 //p3free뤳
 char* Req11(CanHeader* ch,int16_t* p1,int16_t* p2)
 {
-    uint16_t *p3 = Req11r(ch,p1,p2);
-    return ToMb(p3);
+    return ToMb(Req11r(ch,p1,p2));
 }
 
 //cannawcharʸޥХʸˤ
@@ -513,49 +519,43 @@ void dbg_attr(const char* tag,char* a,int len)
 void dbg_str(const char* tag,COMPOSITIONSTRING* cs,int stroffset,
 	     int cloffset,int cllen,
 	     int atoffset,int atlen,
-	     void (*cpy_u16)(uint16_t* dst,const uint16_t* src,int len))
+	     bool han)
 {
-    uint16_t* str = (uint16_t*)((char*)cs+stroffset);
     uint32_t* cl = (uint32_t*)((char*)cs+cloffset);
-    cllen /= 4;
-    uint16_t u16[cl[cllen-1]+1];
-    char ej[sizeof(u16)];
+    char *ej;
     Array lb;
-    int len;
 
+    cllen /= 4;
     ArNew(&lb,1,NULL);
     MSG("\t%s-clause:size %d:%s\n",tag,cllen,ArAdr(Dump4(" %d",cl,cllen,&lb)));
     if(atoffset!=0)
 	dbg_attr(tag,(char*)cs+atoffset,atlen);
+    ArClear(&lb);
     for(int n=0; n<cllen-1; ++n){
-	len = cl[n+1]-cl[n];
-	cpy_u16(u16,str+cl[n],len);
-	ArPrint(&lb,"[%s]",U16ToEj(ej,u16,len));
+	ej = (*WimeData.GetClause)(cs,stroffset,cloffset,n,1);
+	ArPrint(&lb,"[%s]",ej);
+	free(ej);
+    }
+    if(han && ArUsing(&lb)>0){
+	ej = HanToZen(NULL,ArAdr(&lb),-1,false,false);
+	ArPrint(ArClear(&lb),"%s",ej);
+	free(ej);
     }
     MSG("\t%s-str=%s\n",tag,ArAdr(&lb));
     ArDelete(&lb);
 }
 
-void dbg_str_cpy_comp(uint16_t* dst,const uint16_t* src,int len)
-{
-    memcpy(dst,src,len*2);
-}
-
-void dbg_str_cpy_read(uint16_t* dst,const uint16_t* src,int len)
-{
-    SjHan2UniZen(dst,(char*)src,2,0x40,0,len);
-}
-
 void DbgComp(HIMC imc,const char *tag)
 {
     INPUTCONTEXT *ic = (INPUTCONTEXT*)ImmLockIMC(imc);
     COMPOSITIONSTRING *cs = (COMPOSITIONSTRING*)ImmLockIMCC(ic->hCompStr);
 
     MSG("%s:COMPOSITIONSTRING debug\n",tag);
-    dbg_str("comp",cs,cs->dwCompStrOffset,cs->dwCompClauseOffset,cs->dwCompClauseLen,cs->dwCompAttrOffset,cs->dwCompAttrLen,dbg_str_cpy_comp);
-    dbg_str("read",cs,cs->dwCompReadStrOffset,cs->dwCompReadClauseOffset,cs->dwCompReadClauseLen,cs->dwCompReadAttrOffset,cs->dwCompReadAttrLen,dbg_str_cpy_read);
-    dbg_str("result",cs,cs->dwResultStrOffset,cs->dwResultClauseOffset,cs->dwResultClauseLen,0,0,dbg_str_cpy_comp);
-    dbg_str("result-read",cs,cs->dwResultReadStrOffset,cs->dwResultReadClauseOffset,cs->dwResultReadClauseLen,0,0,dbg_str_cpy_read);
+    dbg_str("comp",cs,cs->dwCompStrOffset,cs->dwCompClauseOffset,cs->dwCompClauseLen,cs->dwCompAttrOffset,cs->dwCompAttrLen,false);
+    dbg_str("read",cs,cs->dwCompReadStrOffset,cs->dwCompReadClauseOffset,cs->dwCompReadClauseLen,cs->dwCompReadAttrOffset,cs->dwCompReadAttrLen,true);
+
+    dbg_str("result",cs,cs->dwResultStrOffset,cs->dwResultClauseOffset,cs->dwResultClauseLen,0,0,false);
+    dbg_str("result-read",cs,cs->dwResultReadStrOffset,cs->dwResultReadClauseOffset,cs->dwResultReadClauseLen,0,0,true);
     MSG("\tcursor pos=%d  delta start=%d\n",cs->dwCursorPos,cs->dwDeltaStart);
 
     ImmUnlockIMCC(ic->hCompStr);
@@ -611,7 +611,7 @@ int change_attr(int target,char *attr,const int32_t *cls,int clslen)
   ʸѹ
   tn=ʸֹ(0)
 */
-ChangeTargetStatus ChangeTargetClause(HIMC imc,int tn,const CannaContext_t* cx)
+ChangeTargetStatus SetTarget(HIMC imc,int tn,const CannaContext_t* cx)
 {
     int alen,readalen,r0,r1,clslen,readclslen;
     int32_t *cls,*readcls;
@@ -635,7 +635,7 @@ ChangeTargetStatus ChangeTargetClause(HIMC imc,int tn,const CannaContext_t* cx)
     readclslen = cs->dwCompReadClauseLen/4;
     readcls = (int32_t*)((char*)cs+cs->dwCompReadClauseOffset);
 
-    //LOG("befer change\n");DbgComp(imc,__FUNCTION__);
+    //DbgComp(imc,"before");
     if((r0 = change_attr(tn,attr,cls,clslen)))
        r1 = change_attr(tn,readattr,readcls,readclslen);
     ImmUnlockIMCC(ic->hCompStr);
@@ -648,7 +648,7 @@ ChangeTargetStatus ChangeTargetClause(HIMC imc,int tn,const CannaContext_t* cx)
     if(r0<0)
 	LOG("clause %d is current cl.\n",cx->FixedNum+tn);
     if(r0>0 && r1>0){
-	if(!(r0 = ImmSetCompositionStringW(imc,SCS_CHANGEATTR,attr,alen,readattr,readalen)))
+	if(!(r0 = (*WimeData.SetCompStr)(imc,SCS_CHANGEATTR,attr,alen,readattr,readalen)))
 	    MSG("fail ImmSetCompositionStringW\n");
 #if 0
 	else{
@@ -660,73 +660,137 @@ ChangeTargetStatus ChangeTargetClause(HIMC imc,int tn,const CannaContext_t* cx)
 		MSG("fail ImmNotifyIME\n");
 	}
 #endif
+	//DbgComp(imc,"after");
     }
     return (r0!=0 ? ChangeTargetSuccess:ChangeTargetFail);
 }
 
-//ʸcxΥФ¸롣(ǡ¸)
+/* ʸcxΥФ¸롣(eucjp¸)
+   ɤʸѤҤ餬ʤ¸롣
+*/
 void SaveFixedClause(HIMC imc,CannaContext_t* cx)
 {
-    int32_t *cl,*rcl;
-    uint16_t *uni,*yo;
-    INPUTCONTEXT *ic=(INPUTCONTEXT*)ImmLockIMC(imc);
-    COMPOSITIONSTRING *cs=(COMPOSITIONSTRING*)ImmLockIMCC(ic->hCompStr);
+    Array str;
+    int n;
 
-    cl = (int32_t*)((char*)cs+cs->dwResultClauseOffset);
-    uni = (uint16_t*)((char*)cs+cs->dwResultStrOffset);
-    rcl = (int32_t*)((char*)cs+cs->dwResultReadClauseOffset);
-    yo = (uint16_t*)((char*)cs+cs->dwResultReadStrOffset);
-    for(int n=0; n<(int)(cs->dwResultClauseLen)/4-1; ++n){
-	wstr_add_nwcs(&cx->FixedStr,uni+cl[n],cl[n+1]-cl[n]);
-	wstr_add_wc(&cx->FixedStr,0);
-
-	wstr_add_nwcs(&cx->FixedYomi,yo+rcl[n],rcl[n+1]-rcl[n]);
-	wstr_add_wc(&cx->FixedYomi,0);
-
-	++ cx->FixedNum;
+    ArNew(&str,1,NULL);
+    for(n=cx->FixedNum; GetClause(imc,cx,GCS_RESULTSTR,n,n,&str,NULL); ++n){
+	ListInsert(ArExpand(&cx->FixedStr,ArUsing(&str)),-1,ArAdr(&str));
+	GetClause(imc,cx,GCS_RESULTREADSTR,n,n,&str,NULL);
+	ListInsert(ArExpand(&cx->FixedYomi,ArUsing(&str)),-1,ArAdr(&str));
     }
-    ImmUnlockIMCC(ic->hCompStr);
-    ImmUnlockIMC(imc);
+    cx->FixedNum += n;
+    ArDelete(&str);
 }
 
-/*
-  ʸֹnѴʸcejˤ
-  ejΥ֥å=2
-  ejϥꥢ롣̥ʸɲäʤ
-  atnullǤʤ°֤
+/* ʸֹnn_endޤǤʸ°֤(°ʸnΤΤ֤)
+   str->blocksize2ΤȤToWc()롣
+   ȾѥʤѤҤ餬ʤˤ롣
 */
-Array* ClauseStr(HIMC imc,int n,Array *ej,char* at,const CannaContext_t* cx)
+Array* GetClause(HIMC imc,const CannaContext_t* cx,int req,int n,int n_end,Array* str,char* at)
 {
     INPUTCONTEXT *ic=(INPUTCONTEXT*)ImmLockIMC(imc);
     COMPOSITIONSTRING *cs=(COMPOSITIONSTRING*)ImmLockIMCC(ic->hCompStr);
-    uint16_t *uni;
-    char a;
-    int len;
+    char atdum;
+    Array tmpstr,*f;
+    char *ej;
+    int str_ofs,cl_ofs,at_ofs=-1,atlen=-1,clnum;
+    
+    if(at == NULL)
+	at = &atdum;
+    ArNew(&tmpstr,1,NULL);
 
-    if(n < cx->FixedNum){
-	//nϸʸؤƤ
-	uni = StrListNthWc(cx->FixedStr.adr,cx->FixedNum,n);
-	len = WcLen(uni);
-	a = ATTR_FIXEDCONVERTED; //??? ֤󤳤ΤȤ
-    }else{
-	n -= cx->FixedNum; //ʸ᤬Фʸޤƹͤ
-	int32_t* cl = (int32_t*)((char*)cs+cs->dwCompClauseOffset);
-	uni = (uint16_t*)((char*)cs+cs->dwCompStrOffset) + cl[n];
-	len = cl[n+1]-cl[n];
-	a = *((char*)cs+cs->dwCompAttrOffset+cl[n]);
+    switch(req){
+    case GCS_COMPSTR:
+    case GCS_RESULTSTR:
+	f = (Array*)&cx->FixedStr; //!!! constϤɤ褦
+	break;
+    case GCS_COMPREADSTR:
+    case GCS_RESULTREADSTR:
+	f = (Array*)&cx->FixedYomi;
     }
+    for(; n<cx->FixedNum && n<=n_end; ++n){
+	ej = ListInc(ArAdr(f),n);
+	ArAddN(&tmpstr,ej,strlen(ej));
+    }
+    *at = ATTR_FIXEDCONVERTED; //??? ֤󤳤ΤȤ
 
+    ej = NULL;
+    if(n>=cx->FixedNum || n<n_end){
+	switch(req){
+	case GCS_COMPSTR:
+	    str_ofs = cs->dwCompStrOffset;
+	    cl_ofs = cs->dwCompClauseOffset;
+	    at_ofs = cs->dwCompAttrOffset;
+	    clnum = cs->dwCompClauseLen/sizeof(DWORD)-1;
+	    atlen = cs->dwCompAttrLen;
+	    break;
+	case GCS_COMPREADSTR:
+	    str_ofs = cs->dwCompReadStrOffset;
+	    cl_ofs = cs->dwCompReadClauseOffset;
+	    at_ofs = cs->dwCompReadAttrOffset;
+	    clnum = cs->dwCompReadClauseLen/sizeof(DWORD)-1;
+	    break;
+	case GCS_RESULTSTR:
+	    str_ofs = cs->dwResultStrOffset;
+	    cl_ofs = cs->dwResultClauseOffset;
+	    clnum = cs->dwResultClauseLen/sizeof(DWORD)-1;
+	    break;
+	case GCS_RESULTREADSTR:
+	    str_ofs = cs->dwResultReadStrOffset;
+	    cl_ofs = cs->dwResultReadClauseOffset;
+	    clnum = cs->dwResultReadClauseLen/sizeof(DWORD)-1;
+	}
+	n -= cx->FixedNum;
+	if(n==0 && n_end<=0 && clnum<=0 && atlen>0){
+	    /*!!!
+	      vje delta2.5ǤϡǤʸϣΤޤޤǹʤ
+	      CompAttrLenΤϤ˱ä롣ʸΤΤCompSträƤ롣
+	      apiȤʸ뤬ʸۤŪʸapiǼʤФʤʤ
+	      csľܻȤäƤξ򤤤äڤ뤬apiȤ٤
+	      Ȥꤢ"ʸʤ&&°󤢤"ΤȤΤCompStrΤapiǼ뤳Ȥˤ롣
+	    */
+	    ej = (*WimeData.GetCompStr)(imc,req);
+	    *at = ATTR_INPUT;
+	}else if(n>=clnum || n_end>=clnum){
+	    str = NULL; //ʸֹ椬ϰϳ
+	}else{
+	    if(n_end < 0)
+		n_end = clnum-1;
+	    ej = (*WimeData.GetClause)(cs,str_ofs,cl_ofs,n,n_end-n+1);
+	    if(at_ofs != -1){
+		const int32_t *cl = (typeof(cl))((const char*)cs + cl_ofs);
+		*at = *((char*)cs + at_ofs + cl[n]);
+	    }else
+		*at = ATTR_FIXEDCONVERTED;
+	}
+    }
     if(ej != NULL){
-	U16ToCej(ArAlloc(ej,len+1),uni,len);
-	ArDec(ej);	
-	//Array x,y;ArNew(&x,1,NULL);ArNew(&y,1,NULL);Dump1(" %02x",uni,len*2,&x);Dump1(" %02x",ArAdr(ej),ArUsingBytes(ej),&y);printf("u:%s --> e:%s\n",ArAdr(&x),ArAdr(&y));
+	ArAddN(&tmpstr,ej,strlen(ej));
+	free(ej);
     }
-    if(at != NULL)
-	*at = a;
 
+    if(str != NULL){
+	if(req==GCS_COMPREADSTR || req==GCS_RESULTREADSTR){
+	    char *z = HanToZen(NULL,ArAdr(&tmpstr),ArUsing(&tmpstr),true,true);
+	    ArAddN(ArClear(&tmpstr),z,strlen(z));
+	    free(z);
+	}
+	ArAdd1(&tmpstr,0); //̥ʸ
+	ArClear(str);
+	if(ArBlockSize(str) == 1)
+	    ArSwap(str,&tmpstr);
+	else{
+	    uint16_t *w = ToWc(NULL,ArAdr(&tmpstr));
+	    ArAddN(str,w,WcLen(w)+1);
+	    free(w);
+	}
+    }
+
+    ArDelete(&tmpstr);
     ImmUnlockIMCC(ic->hCompStr);
     ImmUnlockIMC(imc);
-    return ej;
+    return str;
 }
 
 //ʸ֤
@@ -742,66 +806,6 @@ int ClauseLen(HIMC imc,const CannaContext_t* cx)
     return cx->FixedNum + len;
 }
 
-/*
-  n1ʸᤫn2ʸޤǤɤʸeucjp֤
-  e->blocksize == 2
-  eϥꥢ롣
-  e˥̥ʸϤĤuseˤϥȤʤ
-*/
-Array* GetInputClause(HIMC imc,int n1,int n2,const CannaContext_t* cx,Array* e)
-{
-    INPUTCONTEXT *ic = (INPUTCONTEXT*)ImmLockIMC(imc);
-    COMPOSITIONSTRING *cs = (COMPOSITIONSTRING*)ImmLockIMCC(ic->hCompStr);
-    Array src;
-
-    ArNew(&src,2,NULL);
-    while(n1 < min(cx->FixedNum,n2+1)){
-	wstr_add_wcs(&src,StrListNthWc(cx->FixedYomi.adr,cx->FixedNum,n1++));
-    }
-    if(n2 >= cx->FixedNum){
-	n1 -= cx->FixedNum;
-	n2 -= cx->FixedNum;
-	if(n1 <= n2){
-	    int32_t* rcl = (int32_t*)((char*)cs+cs->dwCompReadClauseOffset);
-	    uint16_t* han = (uint16_t*)((char*)cs+cs->dwCompReadStrOffset)+rcl[n1];
-	    wstr_add_nwcs(&src,han,rcl[n2+1]-rcl[n1]);
-	}
-    }
-    wstr_add_wc(&src,0); //̥ʸɲäƤ
-    SjHan2UniZen(ArAdr(&src),ArAdr(&src),2,0x40,1,INT_MAX);
-    ArAlloc(e,WcLen(ArAdr(&src))+1);
-    ArDec(e); //̥ʸϥȤʤ
-    U16ToEj(ArAdr(e),ArAdr(&src),ArUsing(e));
-    *(uint16_t*)ArElem(e,ArUsing(e)) = 0; //U16ToEjΥ̥ʸ8bitʤΤ16bitΥ̥ʸ֤
-
-    ArDelete(&src);
-    ImmUnlockIMCC(ic->hCompStr);
-    ImmUnlockIMC(imc);
-    return e;
-}
-
-/*
-  ɤʸΤʸ֤
-*/
-int InputLen(HIMC imc,CannaContext_t* cx)
-{
-    INPUTCONTEXT *ic = (INPUTCONTEXT*)ImmLockIMC(imc);
-    COMPOSITIONSTRING *cs = (COMPOSITIONSTRING*)ImmLockIMCC(ic->hCompStr);
-    int32_t* rcl = (int32_t*)((char*)cs+cs->dwCompReadClauseOffset);
-    int hanlen = rcl[cs->dwCompReadClauseLen/4-1];
-    uint16_t uzen[hanlen+1];
-    int len=0;
-
-    for(int n=0; n<cx->FixedNum; ++n)
-	len += WcLen(StrListNthWc(cx->FixedYomi.adr,cx->FixedNum,n));
-    SjHan2UniZen(uzen,(char*)cs+cs->dwCompReadStrOffset,2,0x40,1,hanlen);
-    len += WcLen(uzen);
-
-    ImmUnlockIMCC(ic->hCompStr);
-    ImmUnlockIMC(imc);
-    return len;
-}
-
 /*
   °ʸֹ֤
   ʸ̵뤹
@@ -826,15 +830,6 @@ int GetAttrCl(HIMC imc,char at,const CannaContext_t* cx)
     return num<cllen ? cx->FixedNum+num : -1;
 }
 
-#if 0
-int WcToMb(const uint16_t* w,int wlen,char* m,int mlen)
-{
-    int sz = WideCharToMultiByte(CP_ACP,0,(LPCWSTR)w,wlen,m,mlen,NULL,NULL);
-    m[sz] = 0;
-    return sz;
-}
-#endif
-
 /*
   (clpos_b<=ʸֹ<clpos_e)ޤǤCOMPOSITIONSTRINGscsɲä
   ʸֹimcǤο͡ cx->FixedNumϹθƤʤΤǤ餫Ƥȡ
@@ -950,6 +945,6 @@ void CompDelete(Array scs[])
 char GetAttr(HIMC imc,int cl,const CannaContext_t* cx)
 {
     char a;
-    ClauseStr(imc,cl,NULL,&a,cx);
+    GetClause(imc,cx,GCS_COMPSTR,cl,cl,NULL,&a);
     return a;
 }
