@@ -44,7 +44,7 @@ static int total_size(const char* l)
 }
 
 //nܤǤ
-static char* move_to(char* l,int n)
+char* ListInc(char* l,int n)
 {
     while(--n >= 0)
 	l += strlen(l)+1;
@@ -57,8 +57,8 @@ static char* move_to(char* l,int n)
 */
 int ListRemove(char* l,int n)
 {
-    l = move_to(l,n);
-    char *ls = move_to(l,1); //(n+1)ܤ
+    l = ListInc(l,n);
+    char *ls = ListInc(l,1); //(n+1)ܤ
     memcpy(l,ls,total_size(ls));
     return ls-l;
 }
@@ -74,9 +74,13 @@ int ListFind(const char* l,const char* x)
 }
 
 //ꥹlΰposx
+//pos==-1ΤȤꥹȤκǸɲä
 char* ListInsert(char* l0,int pos,const char* x)
 {
-    char *l = move_to(l0,pos);
+    if(pos < 0)
+	pos = ListCount(l0);
+
+    char *l = ListInc(l0,pos);
     int sz = strlen(x)+1;
     memcpy(memmove(l+sz,l,total_size(l)),x,sz);
     return l0;
