@@ -1,3 +1,4 @@
+#define _GNU_SOURCE
 #include <limits.h>
 #include <stdint.h>
 #include <stdio.h>
@@ -8,35 +9,49 @@
 #include <stdbool.h>
 #include "ut.h"
 
+//e-a2a2 s-81a0 u-25a1 t-e296a1 ''
+#define TOFU_E	0xa2a2
+#define TOFU_S	0xa081
+#define TOFU_U2	0xa125
+#define TOFU_U8	0xa196e2
+
 int Verbose;
 
-enum{ EU16_08,EU16_13,U16E_08,U16E_13,EU8_08,EU8_13,
+enum{ EU16_08,EU16_13,
+      U16E_08,U16E_13,
+      EU8_08,EU8_13,
+      U16S,U16S_13,
+      SU16,SU16_13,
       U16U8,CNVMAX};
 iconv_t Icv[CNVMAX];
 
 __attribute__((constructor))
 void iconv_ini(void)
 {
-    const char UTF16[]="utf-16le",UTF8[]="utf-8",
-	       EUCJP[]="euc-jp",EUCX0213[]="euc-jisx0213";
+    const char UTF16[] = "utf-16le",UTF8[]="utf-8",
+	EUCJP[]="euc-jp",EUCX0213[]="euc-jisx0213",
+	SJIS[]="shift_jis",SJIS13[]="shift_jisx0213";
     const char *cs[][2]={ /* to,from */
-	{UTF16,EUCJP},
-	{UTF16,EUCX0213},
-	{EUCJP,UTF16},
-	{EUCX0213,UTF16},
-	{UTF8,EUCJP},
-	{UTF8,EUCX0213},
+	{UTF16,EUCJP},		{UTF16,EUCX0213},
+	{EUCJP,UTF16},		{EUCX0213,UTF16},
+	{UTF8,EUCJP},		{UTF8,EUCX0213},
+	{SJIS,UTF16},		{SJIS13,UTF16},
+	{UTF16,SJIS},		{UTF16,SJIS13},
 	{UTF8,UTF16},
     };
     for(unsigned n=0; n<ITEMS(Icv); ++n)
-	Icv[n] = iconv_open(cs[n][0],cs[n][1]);
+	if(cs[n][0]!=NULL)
+	    Icv[n] = iconv_open(cs[n][0],cs[n][1]);
 }
 
 __attribute__((destructor))
 void iconv_fin(void)
 {
     for(unsigned n=0; n<ITEMS(Icv); ++n)
-	iconv_close(Icv[n]);
+	if(Icv[n] != NULL){
+	    iconv_close(Icv[n]);
+	    Icv[n] = NULL;
+	}
 }
 
 uint16_t* Swap2p(void *x,int dist)
@@ -81,6 +96,10 @@ int* RevInt(int* x)
 int WcLen(const uint16_t* s)
 {
     const uint16_t* s0 = s;
+
+    if(s == NULL)
+	return -1;
+
     while(*(s++) != 0)
 	;
     return s-s0-1;
@@ -119,25 +138,27 @@ char* ToMb(const uint16_t* src)
     Array dst;
     
     ArNew(&dst,1,NULL);
-    for(; *src!=0; ++src){
-	if((*src & 0xff) != 0){
-	    if((*src>>8) & 0x80){
-		*(uint16_t*)ArExpand(&dst,2) = *src;
-	    }else{ //jisx0212
-		*(dp = ArExpand(&dst,3)) = 0x8f;
-		*(uint16_t*)(dp+1) = *src|0x8000;
+    if(src != NULL){
+	for(; *src!=0; ++src){
+	    if((*src & 0xff) != 0){
+		if((*src>>8) & 0x80){
+		    *(uint16_t*)ArExpand(&dst,2) = *src;
+		}else{ //jisx0212
+		    *(dp = ArExpand(&dst,3)) = 0x8f;
+		    *(uint16_t*)(dp+1) = *src|0x8000;
+		}
+	    }else{
+		uc = (unsigned char)(*src>>8);
+		if(uc>=0xa1 && uc<=0xdf){ //Ⱦѥ
+		    dp = ArExpand(&dst,2);
+		    *(dp++) = 0x8e;
+		}else
+		    dp = ArExpand(&dst,1);
+		*dp = uc;
 	    }
-	}else{
-	    uc = (unsigned char)(*src>>8);
-	    if(uc>=0xa1 && uc<=0xdf){ //Ⱦѥ
-		dp = ArExpand(&dst,2);
-		*(dp++) = 0x8e;
-	    }else
-		dp = ArExpand(&dst,1);
-	    *dp = uc;
 	}
+	*(char*)ArExpand(&dst,1) = 0;
     }
-    *(char*)ArExpand(&dst,1) = 0;
     return ArAdr(&dst);
 }
 
@@ -274,133 +295,105 @@ char* Zen2Roman(char* dest,const char* ej0)
 }
 
 /*
-  sjisȾѥʤutf-16ѤҤ餬ʤľ
-  ssz=srcΣʸΥХȿ
-  opt=src˲û
-  syn=
-  ͡ȾѤǻȤäʸιǣʸȤȤ⤢
-  dstsrcƱǤokssz==1ΤȤϼΥХȤ񤭤롣
+  eucjpȾѥ --> ѤҤ餬
+  syn=true
+  zen_asc=asciiѤˤȤtrue
+  ͡dst
+	use_src=ȤäХȿιǣʸȤȤ⤢
+  dstˤϺ磳Хɬ
 */
-int sj_han_to_u16_zen(uint16_t* dst,const char* src,int ssz,unsigned char opt,int syn)
+char* EjHan2Zen(char* dst,const char* src0,int* use_src,bool syn,bool zen_asc)
 {
-    static uint16_t as20[]={
-	0x3000/* */,	0xff01/*!*/,	0x201d/*"*/,	0xff03/*#*/,	/*20*/
-	0xff04/*$*/,	0xff05/*%*/,	0xff06/*&*/,	0x2019/*'*/,	/*24*/
-	0xff08/*(*/,	0xff09/*)*/,	0xff0a/***/,	0xff0b/*+*/,	/*28*/
-	0xff0c/*,*/,	0x2212/*-*/,	0xff0e/*.*/,	0xff0f/*/*/,	/*2c*/
-	0xff10/*0*/,	0xff11/*1*/,	0xff12/*2*/,	0xff13/*3*/,	/*30*/
-	0xff14/*4*/,	0xff15/*5*/,	0xff16/*6*/,	0xff17/*7*/,	/*34*/
-	0xff18/*8*/,	0xff19/*9*/,	0xff1a/*:*/,	0xff1b/*;*/,	/*38*/
-	0xff1c/*<*/,	0xff1d/*=*/,	0xff1e/*>*/,	0xff1f/*?*/,	/*3c*/
-	0xff20/*@*/,	0xff21/*A*/,	0xff22/*B*/,	0xff23/*C*/,	/*40*/
-	0xff24/*D*/,	0xff25/*E*/,	0xff26/*F*/,	0xff27/*G*/,	/*44*/
-	0xff28/*H*/,	0xff29/*I*/,	0xff2a/*J*/,	0xff2b/*K*/,	/*48*/
-	0xff2c/*L*/,	0xff2d/*M*/,	0xff2e/*N*/,	0xff2f/*O*/,	/*4c*/
-	0xff30/*P*/,	0xff31/*Q*/,	0xff32/*R*/,	0xff33/*S*/,	/*50*/
-	0xff34/*T*/,	0xff35/*U*/,	0xff36/*V*/,	0xff37/*W*/,	/*54*/
-	0xff38/*X*/,	0xff39/*Y*/,	0xff3a/*Z*/,	0xff3b/*[*/,	/*58*/
-	0xff3c/*bs*/,	0xff3d/*]*/,	0xff3e/*^*/,	0xff3f/*_*/,	/*5c*/
-	0xff40/*`*/,	0xff41/*a*/,	0xff42/*b*/,	0xff43/*c*/,	/*60*/
-	0xff44/*d*/,	0xff45/*e*/,	0xff46/*f*/,	0xff47/*g*/,	/*64*/
-	0xff48/*h*/,	0xff49/*i*/,	0xff4a/*j*/,	0xff4b/*k*/,	/*68*/
-	0xff4c/*l*/,	0xff4d/*m*/,	0xff4e/*n*/,	0xff4f/*o*/,	/*6c*/
-	0xff50/*p*/,	0xff51/*q*/,	0xff52/*r*/,	0xff53/*s*/,	/*70*/
-	0xff54/*t*/,	0xff55/*u*/,	0xff56/*v*/,	0xff57/*w*/,	/*74*/
-	0xff58/*x*/,	0xff59/*y*/,	0xff5a/*z*/,	0xff5b/*{*/,	/*78*/
-	0xff5c/*|*/,	0xff5d/*}*/,	0x301c/*~*/,	0x3000		/*7c*/
+    static const char as20[]=
+	"ɡ"	"ʡˡܡݡ"	/*20*/
+	""	"䡩"	/*30*/
+	"£ãģţƣ"	"ȣɣʣˣ̣ͣΣ"	/*40*/
+	"Уѣңӣԣգ֣"	"أ٣ڡΡϡ"	/*50*/
+	"ƣ"	""	/*60*/
+	""	"Сáѡ"	/*70*/
+	;
+    static const char wk0[]=
+	"֡ס"		""	//a1
+	""	""	//b0
+	"ĤƤȤʤˤ"	"ͤΤϤҤդؤۤ"	//c0
+	"ߤ"	"󡫡"	//d0
+	;
+    static const char wd0[]=
+	""		""	//a1
+	""	""	//b0
+	"¤ŤǤɢ"	"ФӤ֤٤ܢ"	//c0
+	""	""	//d0
+	;
+    static const char wp0[]={
+	""		""	//a1
+	""	""	//b0
+	""	"ѤԤפڤݢ"	//c0
+	""	""	//d0
     };
-    static uint16_t wk[]={
-	0x3002/**/,	0x300c/**/,	0x300d/**/,	0x3001/**/,	/*0*/
-	0x30fb/**/,	0x3092/**/,	0x3041/**/,	0x3043/**/,	/*1*/
-	0x3045/**/,	0x3047/**/,	0x3049/**/,	0x3083/**/,	/*2*/
-	0x3085/**/,	0x3087/**/,	0x3063/**/,	0x30fc/**/,	/*3*/
-	0x3042/**/,	0x3044/**/,	0x3046/**/,	0x3048/**/,	/*4*/
-	0x304a/**/,	0x304b/**/,	0x304d/**/,	0x304f/**/,	/*5*/
-	0x3051/**/,	0x3053/**/,	0x3055/**/,	0x3057/**/,	/*6*/
-	0x3059/**/,	0x305b/**/,	0x305d/**/,	0x305f/**/,	/*7*/
-	0x3061/**/,	0x3064/**/,	0x3066/**/,	0x3068/**/,	/*8*/
-	0x306a/**/,	0x306b/**/,	0x306c/**/,	0x306d/**/,	/*9*/
-	0x306e/**/,	0x306f/**/,	0x3072/**/,	0x3075/**/,	/*10*/
-	0x3078/**/,	0x307b/**/,	0x307e/**/,	0x307f/**/,	/*11*/
-	0x3080/**/,	0x3081/**/,	0x3082/**/,	0x3084/**/,	/*12*/
-	0x3086/**/,	0x3088/**/,	0x3089/**/,	0x308a/**/,	/*13*/
-	0x308b/**/,	0x308c/**/,	0x308d/**/,	0x308f/**/,	/*14*/
-	0x3093/**/,	0x309b/**/,	0x309c/**/
-    };
-    static uint16_t wd[]={
-	0/**/,	0/**/,	0/**/,	0/**/,	/*0*/
-	0/**/,	0/**/,	0/**/,	0/**/,	/*1*/
-	0/**/,	0/**/,	0/**/,	0/**/,	/*2*/
-	0/**/,	0/**/,	0/**/,	0/**/,	/*3*/
-	0/**/,	0/**/,	0/**/,	0/**/,	/*4*/
-	0/**/,	0x304c/**/,	0x304e/**/,	0x3050/**/,	/*5*/
-	0x3052/**/,	0x3054/**/,	0x3056/**/,	0x3058/**/,	/*6*/
-	0x305a/**/,	0x305c/**/,	0x305e/**/,	0x3060/**/,	/*7*/
-	0x3062/**/,	0x3065/**/,	0x3067/**/,	0x3069/**/,	/*8*/
-	0/**/,	0/**/,	0/**/,	0/**/,	/*9*/
-	0/**/,	0x3070/**/,	0x3073/**/,	0x3076/**/,	/*10*/
-	0x3079/**/,	0x307c/**/,	0/**/,	0/**/,	/*11*/
-	0/**/,	0/**/,	0/**/,	0/**/,	/*12*/
-	0/**/,	0/**/,	0/**/,	0/**/,	/*13*/
-	0/**/,	0/**/,	0/**/,	0/**/,	/*14*/
-	0/**/,	0/**/,	0/**/
-    };
-    static uint16_t wp[]={
-	0/**/,	0/**/,	0/**/,	0/**/,	/*0*/
-	0/**/,	0/**/,	0/**/,	0/**/,	/*1*/
-	0/**/,	0/**/,	0/**/,	0/**/,	/*2*/
-	0/**/,	0/**/,	0/**/,	0/**/,	/*3*/
-	0/**/,	0/**/,	0/**/,	0/**/,	/*4*/
-	0/**/,	0/**/,	0/**/,	0/**/,	/*5*/
-	0/**/,	0/**/,	0/**/,	0/**/,	/*6*/
-	0/**/,	0/**/,	0/**/,	0/**/,	/*7*/
-	0/**/,	0/**/,	0/**/,	0/**/,	/*8*/
-	0/**/,	0/**/,	0/**/,	0/**/,	/*9*/
-	0/**/,	0x3071/**/,	0x3074/**/,	0x3077/**/,
-	0x307a/**/,	0x307d/**/,	0/**/,	0/**/,	/*11*/
-	0/**/,	0/**/,	0/**/,	0/**/,	/*12*/
-	0/**/,	0/**/,	0/**/,	0/**/,	/*13*/
-	0/**/,	0/**/,	0/**/,	0/**/,	/*14*/
-	0/**/,	0/**/,	0/**/
-    };
-    static uint16_t *wdp[]={wd,wp};
+    static const uint16_t *wk=(const uint16_t*)wk0;
+    static const uint16_t *wdp[]={(const uint16_t*)wd0,(const uint16_t*)wp0};
 
-    char cc,nc;
-    int len=1;
-    uint16_t z;
+    const unsigned char *src = (const unsigned char*)src0;
+    int ofs;
+    uint16_t z,tofu='';
+    unsigned char dp;
 
-    if(*src>0 && *(src+1)==0){
-	z = as20[*src-0x20];
-    }else{
-	cc = *src+opt - 0xa1;
-	nc = *(src+ssz)+opt - 0xde;
-	if(syn && (nc==0 || nc==1) && (z=wdp[nc][cc])!=0){
-	    ++len;
+    switch(*src){
+    case 0 ... 0x7f:
+	if(zen_asc){
+	    *(uint16_t*)dst = *(uint16_t*)(as20 + (*(src++)-0x20)*2);
+	    dst += 2;
 	}else
-	    z = wk[cc];
+	    *(dst++) = *(src++);
+	break;
+    case 0x8e:  //ȾѥʤΥ
+	++src;
+	ofs = *(src++)-0xa1;
+	if(syn && src[0]==0x8e && (dp=src[1]-0xde,dp<2) && (z=wdp[dp][ofs])!=tofu){
+	    *(uint16_t*)dst = z;
+	    src += 2;
+	}else{
+	    *(uint16_t*)dst = wk[ofs];
+	}
+	dst += 2;
+	break;
+    case 0x8f:
+	*(dst++) = *(src++);
+    default:
+	*(uint16_t*)dst = *(uint16_t*)src;
+	dst += 2;
+	src += 2;
     }
-    *dst = z;
-    return len;
+    *use_src = src - (const unsigned char*)src0;
+    return dst;
 }
 
 /*
-  sjisȾѥʸutf-16ѤҤ餬ʸľ
-  srcΥ̥ʸlenʸ(Ⱦѥʤǿ)ޤǡ
-  dstκǸ˥̥ʸɲä롣
-  ͡ѤȾѥʸ
-  dstsrcƱǤokssz==1ΤȤϤޤȤưʤ
- */
-int SjHan2UniZen(uint16_t* dst,const char* src,int ssz,unsigned char opt,int syn,int len)
+  eucjpȾѥ --> ѤҤ餬
+  srclen=Хȿ
+  syn=true
+  zen_asc=asciiѤˤȤtrue
+  asciiʸˤ뤿dstsrcĹʤǽ롣
+*/
+char* HanToZen(char* dst,const char* src,int srclen,bool syn,bool zen_asc)
 {
-    int n;
-    const char *src0=src;
-    while(len>0 && *src!=0){
-	n = sj_han_to_u16_zen(dst++,src,ssz,opt,syn);
-	len -= n;
-	src += n*ssz;
+    char *dst0;
+    int use_src;
+
+    if(dst == NULL)
+	dst = malloc(srclen>=0 ?: EjLen(src)*3+1);
+    dst0 = dst;
+
+    if(srclen < 0)
+	srclen = INT_MAX;
+
+    while(*src!=0 && srclen>0){
+	dst = EjHan2Zen(dst,src,&use_src,syn,zen_asc);
+	srclen -= use_src;
+	src += use_src;
     }
     *dst = 0;
-    return (src-src0)/ssz;
+    return dst0;
 }
 
 //eucjpʸƬnʸư롣
@@ -421,9 +414,11 @@ char* ForwardEj(char* ej,int n)
 int EjLen(const char* ej)
 {
     int len=0;
-    while(*ej != 0){
-	ej = ForwardEj((char*)ej,1);
-	++len;
+    if(ej != NULL){
+	while(*ej != 0){
+	    ej = ForwardEj((char*)ej,1);
+	    ++len;
+	}
     }
     return len;
 }
@@ -446,23 +441,24 @@ bool conv(int num,char** in,size_t* ileft,char** out,size_t* oleft)
 	*ileft = ileft0;
 	st = false;
     }
-    if(!st){
+    if(!st && Icv[num+1]!=NULL){
 	st = (iconv(Icv[num+1],in,ileft,out,oleft)!=(size_t)-1);
     }
     return st;
 }
 
 //euc-jp --> ucs2
-uint16_t* EjToU16(uint16_t* dst,char* src)
+uint16_t* EjToU16(uint16_t* dst0,const char* src0)
 {
-    uint16_t *dst0;
+    char *src = strdupa(src0);
     size_t ileft=strlen(src),oleft=EjLen(src)*2;
+    uint16_t *dst;
 
-    if(dst == NULL)
-	dst = malloc(oleft+2);
-    dst0 = dst;
+    if(dst0 == NULL)
+	dst0 = malloc(oleft+2);
+    dst = dst0;
     while(!conv(EU16_08,&src,&ileft,(char**)&dst,&oleft)){
-	*(dst++) = 0xa125; //e-a2a2 u-25a1 ''
+	*(dst++) = TOFU_U2;
 	oleft -= 2;
 	if(*src == (char)0x8f){
 	    ++src;
@@ -478,7 +474,7 @@ uint16_t* EjToU16(uint16_t* dst,char* src)
 }
 
 //canna ej --> ucs2
-uint16_t* CejToU16(uint16_t* dst,uint16_t* src)
+uint16_t* CejToU16(uint16_t* dst,const uint16_t* src)
 {
     char *ej;
 
@@ -487,37 +483,46 @@ uint16_t* CejToU16(uint16_t* dst,uint16_t* src)
     return dst;
 }
 
-/* ucs2 -->euc-jp
-   ??? euc-jisx0213''ԣʸλϤ줺in-ptr,ileft롣
-   euc-jpΤȤʤ櫓狼ʤ
-   iconv(1)ǤϤɤäƤ
-*/
-char* U16ToEj(char* dst,uint16_t* src,int src_len)
+//ucs2 --> ej or sj
+char* u16_to_mb(int cv,int tofu,char* dst0,const uint16_t* src0,int src_len)
 {
-    char *dst0;
+    uint16_t *src;
+    char *dst;
     size_t ileft,oleft;
 
     if(src_len < 0)
 	src_len = WcLen(src);
     ileft = src_len*2;
-    oleft = src_len*3;
-    if(dst == NULL)
-	dst = malloc(oleft+1);
-    dst0 = dst;
-    while(!conv(U16E_08,(char**)&src,&ileft,&dst,&oleft)){
-	*(dst++) = 0xa2; *(dst++) = 0xa2; //e-a2a2 u-25a1 ''
+    oleft = src_len*3; //ejѤ¿˳ݤ
+    src = memcpy(alloca(ileft+2),src0,ileft+2);
+    if(dst0 == NULL)
+	dst0 = malloc(oleft+1);
+    dst = dst0;
+    while(!conv(cv,(char**)&src,&ileft,&dst,&oleft)){
+	*(uint16_t*)dst = tofu;
+	dst += 2;
 	oleft -= 2;
 	++src;
 	ileft -= 2;
     }
     *dst = 0;
-    //iconv(Icv[U16E_08],NULL,NULL,NULL,NULL);
-    //iconv(Icv[U16E_13],NULL,NULL,NULL,NULL);
     return dst0;
 }
 
+/* ucs2 -->euc-jp
+   ??? euc-jisx0213''ԣʸλϤ줺in-ptr,ileft롣
+   euc-jpΤȤʤ櫓狼ʤ
+   iconv(1)ǤϤɤäƤ
+   shift-jisx0213ǤƱͤΤȤ롣u16'Ϥ'sj13Ѵin-ptr,ileft
+   ʸʬ줿out-ptr,oleftϣʸʬʤä
+*/
+char* U16ToEj(char* dst0,const uint16_t* src0,int src_len)
+{
+    return u16_to_mb(U16E_08,TOFU_E,dst0,src0,src_len);
+}
+
 //ucs2 --> canna ej
-uint16_t* U16ToCej(uint16_t* dst,uint16_t* src,int src_len)
+uint16_t* U16ToCej(uint16_t* dst,const uint16_t* src,int src_len)
 {
     char *ej;
 
@@ -535,15 +540,14 @@ char* EjToU8(char* dst,const char* src00)
 	return NULL;
 
     size_t ileft=strlen(src00),oleft=EjLen(src00)*8;
-    int tofu = 0xa196e2; //e-a2a2 u-25a1 t-e296a1 ''
     char *dst0,*src0,*src;
 
-    src0 = src = strdup(src00);
+    src0 = src = strdupa(src00);
     if(dst == NULL)
 	dst = malloc(oleft+1);
     dst0 = dst;
     while(!conv(EU8_08,&src,&ileft,&dst,&oleft)){
-	memcpy(dst,&tofu,3);
+	*(int32_t*)dst = TOFU_U8;
 	dst += 3;
 	oleft -= 3;
 	if(*src == (char)0x8f){
@@ -554,15 +558,15 @@ char* EjToU8(char* dst,const char* src00)
 	ileft -= 2;
     }
     *dst = 0;
-    free(src0);
     return dst0;
 }
 
 //utf16 --> utf8
-char* U16ToU8(char* dst,uint16_t* src,int src_len)
+char* U16ToU8(char* dst0,const uint16_t* src0,int src_len)
 {
+    uint16_t *src;
+    char *dst;
     size_t ileft,oleft;
-    char *dst0;
 
     if(src == NULL)
 	return NULL;
@@ -570,17 +574,71 @@ char* U16ToU8(char* dst,uint16_t* src,int src_len)
 	src_len = WcLen(src);
     ileft = src_len*2;
     oleft = src_len*5;
-    if(dst == NULL)
-	dst = malloc(oleft+1);
-    dst0 = dst;
+    src = memcpy(alloca(ileft+2),src0,ileft+2);
+    if(dst0 == NULL)
+	dst0 = malloc(oleft+1);
+    dst = dst0;
     conv(U16U8,(char**)&src,&ileft,&dst,&oleft);
     *dst = 0;
     return dst0;
 }
 
+//utf16 --> shift-jisx0213
+char* U16ToSj(char* out0,const uint16_t* in0,int in_len)
+{
+    return u16_to_mb(U16S,TOFU_S,out0,in0,in_len);
+}
+
+//ej --> sj
+char* EjToSj(char* out,const char* in)
+{
+    uint16_t *u;
+    
+    u = EjToU16(NULL,in);
+    out = U16ToSj(out,u,-1);
+    free(u);
+    return out;
+}
+
+//!!! EjToU16Ȱˤ뤫
+//sjis --> utf16
+uint16_t* SjToU16(uint16_t* out0,const char* in0,size_t ileft)
+{
+    char *in;
+    uint16_t *out;
+    size_t oleft;
+
+    in = strdupa(in0);
+    if(ileft == (size_t)-1)
+	ileft = strlen(in);
+    oleft=ileft*2; //inȾѥʤξoutϣܤˤʤ롣ʸ롩
+
+    if(out0 == NULL)
+	out0 = malloc(oleft+2);
+    out = out0;
+
+    while(!conv(SU16,&in,&ileft,(char**)&out,&oleft)){
+	*(out++) = TOFU_U2;
+	oleft -= 2;
+	in += 2;
+	ileft -= 2;
+    }
+    *out = 0;
+    return out0;
+}
+
+//sjis --> ej
+char* SjToEj(char* out,const char* in,int in_len)
+{
+    uint16_t *u = SjToU16(NULL,in,in_len);
+    out = U16ToEj(out,u,-1);
+    free(u);
+    return out;
+}
+
 /*
-  eucjpѥɤȾѥɤˤ
-  ͡ȾѥɤǤʸ(1 or 2)
+  eucjpѤҤ餬ʤȾѥʤˤ
+  ͡Ϥʸ(1 or 2)
   dst˥̥ʸĤΤ,5Хɬס
 */
 int EjZen2Han(char *dst,const char *src)
@@ -694,50 +752,62 @@ int EjZen2Han(char *dst,const char *src)
     return moji;
 }
 
-//////////////////////////////////////////////////////////////////////////
+/*
+  eucjpѤҤ餬 --> Ⱦѥ
+*/
+char* ZenToHan(char* dst,const char* src)
+{
+    char *dst0;
 
-Array* wstr_add_nwcs(Array* s,const uint16_t *cs,int len)
-{
-    memcpy(ArExpand(s,len),cs,len*s->blocksize);
-    return s;
-}
-Array* wstr_add_wc(Array* s,uint16_t c)
-{
-    return wstr_add_nwcs(s,&c,1);
-}
-Array* wstr_add_wcs(Array* s,const uint16_t* cs)
-{
-    return wstr_add_nwcs(s,cs,WcLen(cs));
-}
-Array* wstr_add_wstr(Array* s,const Array* p)
-{
-    return wstr_add_nwcs(s,p->adr,p->use);
+    if(dst == NULL)
+	dst = malloc(strlen(src)*4+1);
+    dst0 = dst;
+
+    while(*src != 0){
+	dst += EjZen2Han(dst,src)*2;
+	src += 2;
+    }
+    *dst = 0;
+    return dst0;
 }
 
-//asciiϥȥ륨ǥǵϿ
-Array *wstr_add_zen2han_ej(Array* out,const uint16_t* h)
+/*
+  eucjpѤҤ餬 --> ѥ
+  src_len=ʸ
+*/
+char* HiraToKata(char* dst,const char* src,int src_len)
 {
-    int m,n;
-    uint16_t dst[3];
+    char *dst0;
 
-    for(; *h!=0; ++h){
-	m = EjZen2Han((char*)dst,(char*)h);
-	for(n=0; n<m; ++n){
-	    if(dst[n]<0x100)
-		dst[n] = Swap2(dst[n]);
-	    wstr_add_wc(out,dst[n]);
+    if(src_len < 0)
+	src_len = INT_MAX;
+
+    if(dst == NULL)
+	dst = malloc(src_len<INT_MAX ? src_len*3+1 : strlen(src)+1);
+    dst0 = dst;
+
+    while(*src!=0 && src_len>0){
+	switch((uint8_t)*src){
+	case 0 ... 0x7f:
+	    *(dst++) = *(src++);
+	    break;
+	case 0x8f:
+	    memcpy(dst,src,3);
+	    src += 3;
+	    dst += 3;
+	    break;
+	case 0xa4: //Ҥ餬
+	    *(dst++) = 0xa5;
+	    ++src;
+	    *(dst++) = *(src++);
+	    break;
+	default:
+	    *(int16_t*)dst = *(int16_t*)src;
+	    src += 2;
+	    dst += 2;
 	}
+	--src_len;
     }
-    return out;
-}
-Array* wstr_add_hira2kata_ej(Array* s,const uint16_t* h)
-{
-    uint16_t c;
-    for(; *h!=0; ++h){
-	c = Swap2(*h);
-	if(c>=0xa4a1 && c<=0xa4f4)
-	    c += 0x100; //0xa4xx --> 0xa5xx
-	wstr_add_wc(s,Swap2(c));
-    }
-    return s;
+    *dst = 0;
+    return dst0;
 }
